"""
.. topic:: ``ih-plan download``

    A ``ih-plan download`` subcommand.

    See ``ih-plan download --help`` for more details.
"""

from os import path as osp

import click

from infrahouse_toolkit.cli.lib import get_bucket, get_s3_client


@click.command(name="download")
@click.argument("key_name")
@click.argument("plan_file", required=False)
@click.pass_context
def cmd_download(ctx, key_name, plan_file):
    """
    Download a file from an S3 bucket.

    The specified plan file will be downloaded from the S3 bucket (See ih-plan --help)
    and saved in <plan_file>. By default, the destination file will be a basename of the <key_name>.

    """
    s3_client = get_s3_client(role=ctx.obj["aws_assume_role_arn"])
    bucket = ctx.obj["bucket"] or get_bucket(ctx.obj["tf_backend_file"])
    plan_file = plan_file or osp.basename(key_name)
    with open(plan_file, "wb") as f_desc:
        s3_client.download_fileobj(bucket, key_name, f_desc)
    print(f"Successfully downloaded s3://{bucket}/{key_name} and saved in {plan_file}.")



import pytest

from infrahouse_toolkit.terraform import RunResult, TFStatus, parse_comment
from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.exceptions import IHParseError
from infrahouse_toolkit.terraform.status import RunOutput


@pytest.mark.parametrize("comment_text", ["", None, "foo"])
def test_parse_comment_raises(comment_text):
    with pytest.raises(IHParseError):
        parse_comment(comment_text)


@pytest.mark.parametrize(
    "comment_text, status",
    [
        (
            TFStatus(
                TFS3Backend("foo_backet", "path/to/tf.state"),
                success=True,
                run_result=RunResult(1, 1, 1),
                run_output=RunOutput("no stdout", None),
            ).comment,
            TFStatus(
                TFS3Backend("foo_backet", "path/to/tf.state"),
                success=True,
                run_result=RunResult(1, 1, 1),
                run_output=RunOutput("no stdout", None),
            ),
        ),
        (
            TFStatus(
                TFS3Backend("foo_backet", "path/to/tf.state"),
                success=True,
                run_result=RunResult(1, 1, 1),
                run_output=RunOutput(None, None),
            ).comment,
            TFStatus(
                TFS3Backend("foo_backet", "path/to/tf.state"),
                success=True,
                run_result=RunResult(1, 1, 1),
                run_output=RunOutput(None, None),
            ),
        ),
        # lines are terminated by \r\n
        (
            """# **`s3://foo_backet/path/to/tf.state`**
|  Success  |   🟢 Add |   🟡 Change |   🔴 Destroy |
|:---------:|--------:|-----------:|------------:|
|     ✅     |       1 |          1 |           1 |
<details><summary>STDOUT</summary>

```no stdout```

</details>

<details><summary>STDERR</summary>

```no output```

</details>

<details><summary><i>metadata</i></summary>\r
<p>

```eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJzdGRvdXQiOiAibm8gc3Rkb3V0IiwgInN0ZGVyciI6IG51bGwsICJhZGQiOiAxLCAiY2hhbmdlIjogMSwgImRlc3Ryb3kiOiAxfX0=```

</p>
</details>

    """,
            TFStatus(
                TFS3Backend("foo_backet", "path/to/tf.state"),
                success=True,
                run_result=RunResult(1, 1, 1),
                run_output=RunOutput("no stdout", None),
            ),
        ),
    ],
)
def test_parse_comment(comment_text, status):
    parsed_status = parse_comment(comment_text)
    print(f"{parsed_status = }")
    print(f"{status = }")
    assert parsed_status == status



"""Terraform S3 backend"""
from os import environ

from infrahouse_toolkit.terraform.backends.tfbackend import TFBackend


class TFS3Backend(TFBackend):
    """
    :py:class:`TFS3Backend` describes a Terraform state stored in AWS S3.

    :param bucket: AWS S3 bucket that stores the state.
    :type bucket: str
    :param key: Path inside the S3 bucket to a file with the state.
        Something like ``path/to/github-control.state``.
    :type key: str
    """

    def __init__(self, bucket: str, key: str, region: str = None):
        self._bucket = bucket
        self._key = key
        self._region = region

    @property
    def bucket(self):
        """S3 bucket name"""
        return self._bucket

    @property
    def id(self):
        """A backend URL that identifies the state file."""
        return str(self)

    @property
    def key(self):
        """Path to the Terraform state in the S3 bucket."""
        return self._key

    @property
    def region(self):
        """AWS region. Taken from the environment variable ``AWS_DEFAULT_REGION`` if one is defined."""
        try:
            return self._region or environ.get("AWS_DEFAULT_REGION")
        except KeyError:
            return None

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"s3://{self._bucket}/{self._key}"

    def __eq__(self, other):
        return all((self.bucket == other.bucket, self.key == other.key))



"""Top level exceptions.

The exception hierarchy repeats the structure of the infrahouse_toolkit package.
Each module in the package has its own exceptions.py module.
The module exceptions are inherited from the upper module exceptions.

"""


class IHException(Exception):
    """Generic InfraHouse exception"""



"""
Module for :py:class:`TFStatus`, Terraform plan run status class.
"""

import json
import re
from base64 import b64encode
from collections import namedtuple

from tabulate import tabulate

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.terraform.backends.tfbackend import TFBackend

RunResult = namedtuple("RunResult", "add change destroy")
RunOutput = namedtuple("RunOutput", "stdout stderr")


RE_NO_COLOR = r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
"""


def decolor(text: str) -> str:
    """Remove ANSI escape sequences that color console output."""
    if text:
        ansi_escape = re.compile(RE_NO_COLOR, re.VERBOSE)
        return ansi_escape.sub("", text)

    return text


class TFStatus:
    """
    :py:class:`TFStatus` represents a result of a ``terraform plan`` run.
    It includes outputs (both stdout and stderr) and a summary of changes -
    how many resources are going to be created/changed/destroyed.

    Credit for emojis https://emojicombos.com/
    """

    # pylint: disable=too-many-instance-attributes,too-many-arguments
    # Probably counts could be calculated from
    # affected_resources, but it's optional.
    def __init__(
        self,
        backend: TFBackend,
        success: bool,
        run_result: RunResult,
        run_output: RunOutput,
        affected_resources: RunResult = None,
    ):
        self.backend = backend
        self.success = success
        self.add = run_result.add
        self.change = run_result.change
        self.destroy = run_result.destroy
        self.stdout = run_output.stdout
        self.stderr = run_output.stderr
        self.affected_resources = affected_resources

    @property
    def comment(self):
        """Serialize the status as a comment text eligible to be posted on GitHub."""
        return (
            f"\n# State **`{self.backend.id}`**\n"
            + f"## Affected resources counts\n\n{self.summary_counts}\n"
            + (f"## Affected resources by action\n\n{self.summary_resources}\n" if self.affected_resources else "")
            + f"""<details>\n<summary>STDOUT</summary>\n\n```{decolor(self.stdout) or "no output"}```\n</details>\n"""
            + f"""<details>\n<summary>STDERR</summary>\n\n```{decolor(self.stderr) or "no output"}```\n</details>\n"""
            + f"""<details><summary><i>metadata</i></summary>\n<p>\n```{self.metadata}```\n</p></details>"""
        )

    @property
    def metadata(self):
        """
        Produces a base64 encoded string with a dictionary that can be used
        to create the same instance of the class.
        """
        return b64encode(str(self).encode(DEFAULT_OPEN_ENCODING)).decode(DEFAULT_OPEN_ENCODING)

    @property
    def summary_counts(self):
        """
        Credit for tabulate:
        https://stackoverflow.com/questions/9535954/printing-lists-as-tabular-data

        :return: Formatted table.
        """
        rows = [
            [
                "✅" if self.success else "❌",
                self.add if self.success and self.add is not None else "❔",
                self.change if self.success and self.change is not None else "❔",
                self.destroy if self.success and self.destroy is not None else "❔",
            ]
        ]
        return tabulate(
            rows,
            headers=[
                "Success",
                f"{'🟢' if self.success and self.add is not None and self.add > 0 else ''} Add",
                f"{'🟡' if self.success and self.change is not None and self.change > 0 else ''} Change",
                f"{'🔴' if self.success and self.destroy is not None and self.destroy > 0 else ''} Destroy",
            ],
            colalign=("center",),
            tablefmt="pipe",
        )

    @property
    def summary_resources(self):
        """
        Produces a string with a table that lists all added/modified/deleted resources.
        """
        if all(
            (
                self.affected_resources,
                self.affected_resources.add is not None,
                self.affected_resources.change is not None,
                self.affected_resources.destroy is not None,
            )
        ):
            rows = (
                [["🟢", f"`{field}`"] for field in self.affected_resources.add]
                + [["🟡", f"`{field}`"] for field in self.affected_resources.change]
                + [["🔴", f"`{field}`"] for field in self.affected_resources.destroy]
            )
            return (
                tabulate(
                    rows,
                    headers=["Action", "Resources"],
                    colalign=("center",),
                    tablefmt="pipe",
                )
                if rows
                else ""
            )

        return "No affected resources"

    def __eq__(self, other):
        return all(getattr(self, x) == getattr(other, x) for x in self.__dict__ if x != "affected_resources")

    def __repr__(self):
        return json.dumps(
            {
                self.backend.id: {
                    "success": self.success,
                    "stdout": self.stdout,
                    "stderr": self.stderr,
                    "add": self.add,
                    "change": self.change,
                    "destroy": self.destroy,
                }
            }
        )



"""
.. topic:: ``gpg.py``

    GPG helper functions.
"""

from contextlib import contextmanager
from os import path as osp
from subprocess import Popen
from tempfile import NamedTemporaryFile, TemporaryDirectory

from botocore.exceptions import ClientError

from infrahouse_toolkit import DEFAULT_ENCODING, DEFAULT_OPEN_ENCODING, LOG
from infrahouse_toolkit.cli.ih_s3_reprepro.aws import get_client


@contextmanager
def gpg_home() -> str:
    """
    :return: GPG home directory
    """
    with TemporaryDirectory() as path:
        LOG.debug("GPG home is %s", path)
        yield path


@contextmanager
def gpg(secret_key=None, role_arn=None, secret_passphrase=None, region: str = None) -> str:
    """
    Prepare GPG keyring and yield path to it.
    If no function arguments are specified, it will return the default path ~/.gnupg.

    If secret_key is specified, the function will pull the GPG key from this secret.
    Optionally, it will assume a role, if given.

    If secret_passphrase isn't specified, gpg and reprepro will ask a passphrase from a terminal.
    If specified, the function will pull the passprase from AWS secretsmanager secret secret_passphrase,
    save it in a temporary file, and create a GPG config so gpg can read the passphrase from
    the temporary file.

    The GPG private key will be imported. Again, if secret_key is specified.
    So the GPG home will include the private key.

    :param secret_key: AWS secret id (name or ARN) with a GPG private key.
    :type secret_key: str
    :param role_arn: If specified, assume this role in AWS client.
    :type role_arn: str
    :param secret_passphrase: AWS secret id (name or ARN) with a passphrase for the GPG private key.
    :type secret_passphrase: str
    :param region: AWS region name.
    :type region: str
    :return: Path to GPG homedir.
    :rtype: str
    """
    if secret_key:
        LOG.debug("Reading GPG key from %s", secret_key)
        try:
            secrets_manager = get_client("secretsmanager", role_arn=role_arn, region=region)
            key = secrets_manager.get_secret_value(SecretId=secret_key)["SecretString"]
        except ClientError as err:
            LOG.error(err)
            if err.response["Error"]["Code"] == "ResourceNotFoundException":
                LOG.error("Make sure that you set a secret value to the key %s:", secret_key)
                LOG.error("ih-s3-reprepro ... set-secret-value %s /path/to/file/with/%s-value", secret_key, secret_key)
            raise

        with gpg_home() as homedir, NamedTemporaryFile() as gpg_key_desc, NamedTemporaryFile() as gpg_passphrase_desc:
            gpg_key_desc.write(key.encode(DEFAULT_ENCODING))
            gpg_key_desc.flush()

            cmd = ["gpg", "--homedir", homedir]

            if secret_passphrase:
                LOG.debug("Reading GPG key passphrase from %s", secret_passphrase)
                passphrase = secrets_manager.get_secret_value(SecretId=secret_passphrase)["SecretString"]
                gpg_passphrase_desc.write(passphrase.encode(DEFAULT_ENCODING))
                gpg_passphrase_desc.flush()
                write_gpg_cong(
                    osp.join(homedir, "gpg.conf"),
                    {"batch": None, "passphrase-file": gpg_passphrase_desc.name, "pinentry-mode": "loopback"},
                )

            cmd.extend(["--import", gpg_key_desc.name])
            LOG.debug("Executing: %s", " ".join(cmd))
            proc = Popen(cmd)
            proc.communicate()
            yield homedir
    else:
        yield osp.expanduser("~/.gnupg")


def write_gpg_cong(path: str, options: dict):
    """
    Prepare GPG config file.

    :param path: Path name to the GPG config file.
    :type path: str
    :param options: A dictionary with options. For one word options the key value is None.
    :type options: dict
    """
    LOG.debug("Saving GPG config in %s", path)
    with open(path, "w", encoding=DEFAULT_OPEN_ENCODING) as gpg_conf_desc:
        for key, value in options.items():
            gpg_conf_desc.write(f"{key}")
            if value:
                gpg_conf_desc.write(f" {value}")
            gpg_conf_desc.write("\n")



import json

from infrahouse_toolkit.cli.ih_plan.cmd_min_permissions import ActionList


def test_load_from_file(tmpdir):
    existing = tmpdir.join("existing.json")
    existing.write(json.dumps(["foo", "bar"]))
    actions = ActionList()
    actions.load_from_file(str(existing))
    assert actions.actions == ["bar", "foo"]



from infrahouse_toolkit.terraform.backends.s3backend import TFS3Backend


def test_eq():
    assert TFS3Backend("foobucket", "foo.key") == TFS3Backend("foobucket", "foo.key")
    assert TFS3Backend("foobucket", "foo.key") != TFS3Backend("barbucket", "foo.key")



import pytest

from infrahouse_toolkit.terraform.backends import (
    IHUnknownBackend,
    TFS3Backend,
    get_backend,
)


def test_get_backend_raises():
    with pytest.raises(IHUnknownBackend):
        get_backend("foo")


def test_get_backend():
    assert get_backend("s3://foo-bucket/foo.key") == TFS3Backend("foo-bucket", "foo.key")



"""Auxiliary functions for command line tools."""
import boto3
import hcl

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING

DEFAULT_TF_BACKEND_FILE = "terraform.tf"


def get_bucket(tf_file=DEFAULT_TF_BACKEND_FILE) -> str:
    """
    Find bucket name in a Terraform backend configuration.

    :param tf_file: Path to the Terraform backend configuration.
    :type tf_file: str
    :return: Bucket name.
    """
    with open(tf_file, encoding=DEFAULT_OPEN_ENCODING) as f_desc:
        obj = hcl.load(f_desc)
        return obj["terraform"]["backend"]["s3"]["bucket"]


def get_backend_key(tf_file=DEFAULT_TF_BACKEND_FILE) -> str:
    """
    Find terraform state filename in a Terraform backend configuration.

    :param tf_file: Path to the Terraform backend configuration.
    :type tf_file: str
    :return: Path to Terraform state in S3.
    """
    with open(tf_file, encoding=DEFAULT_OPEN_ENCODING) as f_desc:
        obj = hcl.load(f_desc)
        return obj["terraform"]["backend"]["s3"]["key"]


def get_s3_client(role: str = None):
    """
    Get a boto3 S3 client to work with AWS S3.
    If a role is given, assume it.

    :param role: ARN of a role to be assumed
    :return: A boto3 S3 client object
    """
    if role:
        client = boto3.client("sts")
        response = client.assume_role(DurationSeconds=900, RoleArn=role, RoleSessionName="infrahouse-toolkit")
        session = boto3.Session(
            aws_access_key_id=response["Credentials"]["AccessKeyId"],
            aws_secret_access_key=response["Credentials"]["SecretAccessKey"],
            aws_session_token=response["Credentials"]["SessionToken"],
        )
    else:
        session = boto3.Session()

    return session.client("s3")



"""
.. topic:: ``ih-s3-reprepro get-secret-value``

    A ``ih-s3-reprepro get-secret-value`` subcommand.

    See ``ih-s3-reprepro get-secret-value --help`` for more details.
"""
import sys

import click
from botocore.exceptions import BotoCoreError, ClientError
from click import Context

from infrahouse_toolkit import LOG
from infrahouse_toolkit.cli.ih_s3_reprepro.aws import get_client


@click.command(name="get-secret-value")
@click.argument("secret_id")
@click.pass_context
def cmd_get_secret_value(ctx: Context, secret_id):
    """
    Retrieve a secret value.
    """
    role_arn = ctx.parent.params["role_arn"]
    try:
        client = get_client("secretsmanager", role_arn=role_arn, region=ctx.parent.params["aws_region"])
        response = client.get_secret_value(
            SecretId=secret_id,
        )
        LOG.info("%s", response["SecretString"])
        LOG.info("Be careful with copy&pasting the secret value.")
    except (ClientError, BotoCoreError) as err:
        LOG.error(err)
        sys.exit(1)



from os import environ

import pytest

from infrahouse_toolkit.terraform.backends.s3backend import TFS3Backend


@pytest.mark.parametrize(
    "aws_default_region, args, expected",
    [
        (None, ("foobucket", "foo.key"), None),
        ("foo_region", ("foobucket", "foo.key"), "foo_region"),
        ("foo_region", ("foobucket", "foo.key", "bar_region"), "bar_region"),
    ],
)
def test_region(aws_default_region, args, expected):
    if aws_default_region:
        # noinspection PyTypeChecker
        environ["AWS_DEFAULT_REGION"] = aws_default_region
    elif "AWS_DEFAULT_REGION" in environ:
        del environ["AWS_DEFAULT_REGION"]
    assert TFS3Backend(*args).region == expected



from textwrap import dedent

from infrahouse_toolkit.cli.ih_plan.cmd_min_permissions import ActionList


def test_parse_trace(tmpdir):
    tracefile = tmpdir.join("trace")
    tracefile.write(
        dedent(
            """
            {"aws.operation": "PutScalingPolicy","aws.service": "Auto Scaling"}
            {"aws.operation": "DescribePolicies","aws.service": "Auto Scaling"}
            """
        )
    )
    actions = ActionList()
    actions.parse_trace(str(tracefile))
    assert actions.actions == [
        "autoscaling:DescribePolicies",
        "autoscaling:PutScalingPolicy",
    ]
